/*
	Revision history:
	1.0 : initial release
	1.1 : added support for batch jobs (scripting)
	1.2 : optimised blend function
	1.3 : added mask bitmap support and error detection (now requires VD1.4)
*/

#include <windows.h>
#include <commctrl.h>
#include <io.h>

#include <cstdio>

#include "resource.h"
#include "filter.h"
#include "scriptinterpreter.h"
#include "scriptvalue.h"

///////////////////////////////////////////////////////////////////////////

int fblenderRunProc (const FilterActivation *fa, const FilterFunctions *ff);
int fblenderStartProc (FilterActivation *fa, const FilterFunctions *ff);
int fblenderEndProc (FilterActivation *fa, const FilterFunctions *ff);
int fblenderInitProc (FilterActivation *fa, const FilterFunctions *ff);
int fblenderConfigProc (FilterActivation *fa, const FilterFunctions *ff, HWND hwnd);
void fblenderStringProc (const FilterActivation *fa, const FilterFunctions *ff, char *buf);
void fblenderScriptConfigProc (IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);
bool fblenderFssProc (FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);
long fblenderParamProc (FilterActivation *fa, const FilterFunctions *ff);

///////////////////////////////////////////////////////////////////////////

//Accept only V6 filter API (VirtualDub 1.4 or better)
#define VDUB_14 6

enum
{
	TYPE_STORE = 0,
	TYPE_SWAP,
	TYPE_BLEND
};

#define RED GetRValue
#define GREEN GetGValue
#define BLUE GetBValue

ScriptFunctionDef fblender_func_defs[]={
	{ (ScriptFunctionPtr)fblenderScriptConfigProc, "Config", "0iii" },
	{ (ScriptFunctionPtr)fblenderScriptConfigProc, "Config", "0is" },
	{ (ScriptFunctionPtr)fblenderScriptConfigProc, "Config", "0i" },
	{ NULL },
};

CScriptObject fblender_obj={
	NULL, fblender_func_defs
};

typedef struct MyFilterData {
	char			typeID;
} MyFilterData;

struct FilterDefinition filterDef_fblender = {

	NULL, NULL, NULL,				// next, prev, module
	"filter blender (1.3)",			// name
	"Allows creation of two filter paths and blending the results.",
									// desc
	"Steven Don",					// maker
	NULL,							// private_data
	sizeof (MyFilterData),			// inst_data_size

	fblenderInitProc,				// initProc
	NULL,							// deinitProc
	fblenderRunProc,				// runProc
	fblenderParamProc,				// paramProc
	fblenderConfigProc,				// configProc
	fblenderStringProc,				// stringProc
	fblenderStartProc,				// startProc
	fblenderEndProc,				// endProc

	&fblender_obj,					// script_obj
	fblenderFssProc,				// fssProc
};

///////////////////////////////////////////////////////////////////////////

extern "C" int __declspec (dllexport) __cdecl VirtualdubFilterModuleInit2 (FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec (dllexport) __cdecl VirtualdubFilterModuleDeinit (FilterModule *fm, const FilterFunctions *ff);

static FilterDefinition *fd_fblender;

///////////////////////////////////////////////////////////////////////////

int __declspec (dllexport) __cdecl VirtualdubFilterModuleInit2 (FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat)
{
	if (! (fd_fblender = ff->addFilter (fm, &filterDef_fblender, sizeof (FilterDefinition))))
		return 1;

	vdfd_ver = VIRTUALDUB_FILTERDEF_VERSION;
	vdfd_compat = VDUB_14;

	return 0;
}

void __declspec (dllexport) __cdecl VirtualdubFilterModuleDeinit (FilterModule *fm, const FilterFunctions *ff)
{
	ff->removeFilter (fd_fblender);
}

///////////////////////////////////////////////////////////////////////////

//Image dimensions
int width, height;
int imgStride, maskStride;

//Frames used
unsigned char *maskFrame = NULL;
Pixel32 *parkedFrames [2] = { NULL, NULL };

//Lookup tables
unsigned char blendTable [256][256];

//Parameters
int levelA = 50, levelB = 50;
char maskFilename [FILENAME_MAX];
bool usemask = false;

void calcBlendTables ()
{
	//Create blending tables
	if (usemask)
	{
		/*
			Mask based blending:
			a = mask bitmap (alpha value)
			b = source bitmap
		*/
		for (int a = 0; a < 256; ++a)
		{
			for (int b = 0; b < 256; ++b)
			{
				blendTable [a][b] = a * b / 255;
			}
		}
	} else {
		/*
			Percentage based blending:
			a = source bitmap A
			b = source bitmap B
		*/
		for (int a = 0; a < 256; ++a)
		{
			for (int b = 0; b < 256; ++b)
			{
				int result = (a * levelA + b * levelB) / 100;
				if (result > 255) result = 255;
				blendTable [a][b] = (unsigned char)result;
			}
		}
	}
}

bool loadMaskBitmap (const FilterFunctions *ff)
{
	//Load image
	FILE *f = fopen (maskFilename, "rb");
	if (!f)
	{
		ff->Except ("File not found: %s", maskFilename);
		return false;
	}

	try
	{
		BITMAPFILEHEADER bmfh = { 0 };
		BITMAPINFOHEADER bmih = { 0 };

		//Load file header
		fread (&bmfh, sizeof (BITMAPFILEHEADER), 1, f);
		if (bmfh.bfType != MAKEWORD ('B', 'M'))
			throw "Invalid bitmap file header.";

		//Load info header
		fread (&bmih, sizeof (BITMAPINFOHEADER), 1, f);
		if (bmih.biBitCount != 8)
			throw "Not 8bit greyscale image.";
		if (bmih.biCompression != BI_RGB)
			throw "Not raw image.";
		if (bmih.biWidth != width || bmih.biHeight != height)
			throw "Image dimension mismatch.";

		//Allocate memory
		maskStride = (width + 3) & ~3;
		maskFrame = new unsigned char [maskStride * height];
		if (!maskFrame)
			ff->ExceptOutOfMemory ();

		//Skip to actual bitmap data
		fseek (f, bmfh.bfOffBits, SEEK_SET);
		if (fread (maskFrame, maskStride, height, f) != (size_t)height)
			throw "Truncated bitmap file.";

	} catch (const char *e)
	{
		fclose (f);
		delete [] maskFrame;
		maskFrame = NULL;

		ff->Except ("Error loading mask bitmap: %s", e);

		return false;
	}

	//All done
	return true;
}

void blendFrames (Pixel32 *srcA, Pixel32 *srcB, Pixel32 *dst, int size)
{
	while (size--)
	{
		Pixel32 a = *srcA++;
		Pixel32 b = *srcB++;

		*dst++ = RGB (
				blendTable [RED (a)  ][RED (b)  ],
				blendTable [GREEN (a)][GREEN (b)],
				blendTable [BLUE (a) ][BLUE (b) ]
			);
	}
}

void blendMasked (Pixel32 *srcA, Pixel32 *srcB, Pixel32 *dst, int)
{
	//Scanline pointers
	unsigned char *mask = maskFrame;

	//Pixel pointers
	Pixel32 *A;
	Pixel32 *B;
	Pixel32 *D;
	unsigned char *M;

	//Handle scanlines
	for (int y = 0; y < height; ++y)
	{
		//Set up pixel pointers and advance scanline pointers
		A = srcA; srcA += imgStride;
		B = srcB; srcB += imgStride;
		D = dst;  dst  += imgStride;
		M = mask; mask += maskStride;

		//Handle pixels
		for (int x = 0; x < width; ++x)
		{
			Pixel32 a = *A++;
			Pixel32 b = *B++;
			unsigned char m1 = *M++;
			unsigned char m2 = m1 ^ 0xFF;

			*D++ = RGB (
					blendTable [m1][RED (a)  ] + blendTable [m2][RED (b)  ],
					blendTable [m1][GREEN (a)] + blendTable [m2][GREEN (b)],
					blendTable [m1][BLUE (a) ] + blendTable [m2][BLUE (b) ]
				);
		}
	}
}

///////////////////////////////////////////////////////////////////////////

int fblenderStartProc (FilterActivation *fa, const FilterFunctions *ff)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	//Allocate memory
	if (!parkedFrames [0]) parkedFrames [0] = new Pixel32 [fa->src.size / sizeof (Pixel32)];
	if (!parkedFrames [0]) ff->ExceptOutOfMemory ();
	if (!parkedFrames [1]) parkedFrames [1] = new Pixel32 [fa->src.size / sizeof (Pixel32)];
	if (!parkedFrames [1]) ff->ExceptOutOfMemory ();

	//Handle dimensions
	if (!width || !height)
	{
		//Store size
		width = fa->src.w;
		height = fa->src.h;
		//Other data
		imgStride = fa->src.pitch / sizeof (Pixel32);
	} else {
		//Compare size
		if (width != fa->src.w || height != fa->src.h)
		{
			ff->Except ("Size mismatch: All paths should be %dx%d pixels", width, height);
			return 1;
		}
	}

	//Handle blending
	if (mfd->typeID == TYPE_BLEND)
	{
		calcBlendTables ();
		if (usemask && !loadMaskBitmap (ff))
		{
			ff->Except ("Unabled to load mask bitmap.");
			return 2;
		}
	}

	return 0;
}

int fblenderRunProc (const FilterActivation *fa, const FilterFunctions *ff)
{
	//Get instance data
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	switch (mfd->typeID)
	{
	case TYPE_STORE :
		//Store the current frame as the original frame
		memcpy (parkedFrames [0], fa->src.data, fa->src.size);
		break;
	case TYPE_SWAP :
		//Store the current frame as the second frame
		memcpy (parkedFrames [1], fa->src.data, fa->src.size);
		//Put the original frame in its place
		memcpy (fa->dst.data, parkedFrames [0], fa->src.size);
		break;
	case TYPE_BLEND :
		//Is a mask bitmap used?
		(usemask ? blendMasked : blendFrames) (parkedFrames [1], fa->src.data, fa->dst.data, fa->src.size / sizeof (Pixel32));
		break;
	}

	return 0;
}

int fblenderEndProc (FilterActivation *fa, const FilterFunctions *ff)
{
	//Frame size doesn't matter
	width = height = 0;

	//Deallocate memory used
	delete [] maskFrame; maskFrame = NULL;
	delete [] parkedFrames [0]; parkedFrames [0] = NULL;
	delete [] parkedFrames [1]; parkedFrames [1] = NULL;

	//All done
	return 0;
}

int fblenderInitProc (FilterActivation *fa, const FilterFunctions *ff)
{
	//Initialise to default values
	levelA = levelB = 50;
	usemask = false;
	strcpy (maskFilename, "");
	return 0;
}

void EnableItems (HWND hdlg, char typeID)
{
	bool blend = typeID == 2;
	bool mask = SendDlgItemMessage (hdlg, IDC_RADIO5, BM_GETCHECK, 0, 0) == BST_CHECKED;
	bool enabled = blend;

	//Radio buttons
	SendDlgItemMessage (hdlg, IDC_RADIO1, BM_SETCHECK, typeID == 0, 0);
	SendDlgItemMessage (hdlg, IDC_RADIO2, BM_SETCHECK, typeID == 1, 0);
	SendDlgItemMessage (hdlg, IDC_RADIO3, BM_SETCHECK, typeID == 2, 0);

	//Blend panel
	EnableWindow (GetDlgItem (hdlg, IDC_STATIC_BORDER), enabled);
	EnableWindow (GetDlgItem (hdlg, IDC_RADIO4), enabled);
	EnableWindow (GetDlgItem (hdlg, IDC_RADIO5), enabled);

	//Percentage blending
	enabled = blend && !mask;
	SendDlgItemMessage (hdlg, IDC_RADIO4, BM_SETCHECK, enabled, 0);
	EnableWindow (GetDlgItem (hdlg, IDC_STATIC_PB), enabled);
	EnableWindow (GetDlgItem (hdlg, IDC_EDIT1), enabled);
	EnableWindow (GetDlgItem (hdlg, IDC_EDIT2), enabled);
	EnableWindow (GetDlgItem (hdlg, IDC_STATIC_PMA), enabled);
	EnableWindow (GetDlgItem (hdlg, IDC_STATIC_PMB), enabled);

	//Mask blending
	enabled = blend && mask;
	SendDlgItemMessage (hdlg, IDC_RADIO5, BM_SETCHECK, enabled, 0);
	EnableWindow (GetDlgItem (hdlg, IDC_BUTTON_BROWSE), enabled);
}

BOOL CALLBACK fblenderConfigDlgProc (HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
	MyFilterData *mfd = (MyFilterData *)GetWindowLong (hdlg, DWL_USER);

	switch (msg)
	{
	case WM_INITDIALOG:
		SetWindowLong (hdlg, DWL_USER, lParam);
		mfd = (MyFilterData *)lParam;

		SendDlgItemMessage (hdlg, IDC_RADIO5, BM_SETCHECK, usemask, 0);

		EnableItems (hdlg, mfd->typeID);

		//Set up blend percentages
		char leveltext [4];
		sprintf (leveltext, "%d", levelA);
		SetDlgItemText (hdlg, IDC_EDIT1, leveltext);
		sprintf (leveltext, "%d", levelB);
		SetDlgItemText (hdlg, IDC_EDIT2, leveltext);

		//Set up mask bitmap filename
		SetDlgItemText (hdlg, IDC_EDIT3, maskFilename);

		return TRUE;

	case WM_COMMAND:
		switch (LOWORD (wParam))
		{
		case IDC_RADIO1:
			EnableItems (hdlg, mfd->typeID = TYPE_STORE);
			break;
		case IDC_RADIO2:
			EnableItems (hdlg, mfd->typeID = TYPE_SWAP);
			break;
		case IDC_RADIO3:
			EnableItems (hdlg, mfd->typeID = TYPE_BLEND);
			break;
		case IDC_RADIO4:
			EnableItems (hdlg, mfd->typeID);
			break;
		case IDC_RADIO5:
			EnableItems (hdlg, mfd->typeID);
			break;
		case IDC_BUTTON_BROWSE :
			{
				//Set up filename
				char fn [FILENAME_MAX];
				memcpy (fn, maskFilename, sizeof (fn));

				//Get bitmap file path
				OPENFILENAME ofn = { sizeof (OPENFILENAME) };
				ofn.hwndOwner = hdlg;
				ofn.lpstrFilter = "Windows Bitmap File\0*.BMP\0";
				ofn.lpstrFile = fn;
				ofn.nMaxFile = sizeof (fn);
				ofn.lpstrTitle = "Choose mask bitmap";
				ofn.Flags = OFN_FILEMUSTEXIST;
				if (GetOpenFileName (&ofn))
				{
					SetDlgItemText (hdlg, IDC_EDIT3, fn);
				}
			}
			break;
		case IDOK:
			if (mfd->typeID == TYPE_BLEND)
			{
				//Check whether a mask bitmap is in use
				usemask = SendDlgItemMessage (hdlg, IDC_RADIO5, BM_GETCHECK, 0, 0) == BST_CHECKED;

				//Store mask filename
				GetDlgItemText (hdlg, IDC_EDIT3, maskFilename, sizeof (maskFilename));
				if (usemask && _access (maskFilename, 4) != 0)
				{
					MessageBox (hdlg, "Mask bitmap cannot be opened.", NULL, MB_ICONERROR);
					break;
				}

				//Store blend percentages
				levelA = GetDlgItemInt (hdlg, IDC_EDIT1, NULL, FALSE);
				levelB = GetDlgItemInt (hdlg, IDC_EDIT2, NULL, FALSE);
				if (levelA > 300) levelA = 300;
				if (levelB > 300) levelB = 300;

				//Update the blend tables to reflect changes
				calcBlendTables ();
			}
			EndDialog (hdlg, 0);
			return TRUE;
		case IDCANCEL:
			EndDialog (hdlg, 1);
			return FALSE;
		}
		break;
	}

	return FALSE;
}

int fblenderConfigProc (FilterActivation *fa, const FilterFunctions *ff, HWND hwnd)
{
	return DialogBoxParam (fa->filter->module->hInstModule,
		MAKEINTRESOURCE (IDD_DIALOG_CONFIG), hwnd,
		fblenderConfigDlgProc, (LPARAM)fa->filter_data);
}

void fblenderStringProc (const FilterActivation *fa, const FilterFunctions *ff, char *str)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	switch (mfd->typeID)
	{
	case TYPE_STORE:
		sprintf (str, " (begin path A)");
		break;
	case TYPE_SWAP:
		sprintf (str, " (begin path B)");
		break;
	case TYPE_BLEND:
		if (usemask)
			sprintf (str, " (blend with %s)", maskFilename);
		else
			sprintf (str, " (blend A@%d%%, B@%d%%)", levelA, levelB);
		break;
	}
}

void fblenderScriptConfigProc (IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc)
{
	FilterActivation *fa = (FilterActivation *)lpVoid;
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	mfd->typeID = argv [0].asInt ();
	if (mfd->typeID == TYPE_BLEND)
	{
		if (argc == 3)
		{
			//Percentages
			levelA = argv [1].asInt ();
			levelB = argv [2].asInt ();
			usemask = false;
		} else if (argc == 2 && argv [1].isString ())
		{
			//Mask bitmap filename
			int size = sizeof (maskFilename);
			strncpy (maskFilename, *argv [1].asString (), size) [size - 1] = 0;
			usemask = true;
		}
	}
}

bool fblenderFssProc (FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	if (mfd->typeID == TYPE_BLEND)
	{
		if (usemask)
		{
			//Why doesn't the CRT have a function to escape backslashes?
			char temp [FILENAME_MAX * 4];
			char *src = maskFilename;
			char *tgt = temp;
			do
			{
				*tgt++ = *src;
				if (*src == '\\')
					*tgt++ = '\\';
			} while (*src++);
			_snprintf (buf, buflen, "Config (%d, \"%s\")", mfd->typeID, temp);
		} else
			_snprintf (buf, buflen, "Config (%d, %d, %d)", mfd->typeID, levelA, levelB);
	} else {
		_snprintf (buf, buflen, "Config (%d)", mfd->typeID);
	}
	return true;
}

long fblenderParamProc (FilterActivation *fa, const FilterFunctions *ff)
{
	return 0;
}